<?php
/**
 * @file
 * classes for entity object, not depending on LS.
 */

/**
 * Class managing building dynamic entities.
 *
 * Building entity_info, entity_property_info, schema. Take care of Drupal
 * fields and custom fields.
 */
class CustomizableEntitiesEntityInfo {
  // store the class info name into entity_info().
  /*
   * Each class that extend the CustomizableEntitiesEntityInfo class must
   * provide at least this parameter : the entity machine name. It must be a
   * string, not empty, with no special charaters (underscore is allowed).
   */
  protected $entityname = NULL;
  /*
   * String, default label for the entity. If not provided, it will be the
   * entity machine name.
   */
  protected $label = NULL;
  /*
   * String, default plural label for the entity. If not provided, it will be the
   * entity label.
   */
  protected $plural_label = NULL;
  /*
   * String, module that declare this entity. The customizable_entities_build_entity_infos()
   * takes care of setting the module based on the hook_customizable_entities_api()
   * function.
   */
  protected $module = NULL;

  // Add custom keys to the entity info datas :
  // And the prefix root path for view, edit, ...:
  protected $crud_path = '';
  // Create an overview page on the path.
  protected $overview_path;

  // Uri for the overview page.
  protected $root_admin_ui = '';
  // If the entity manage revision.
  protected $revision = NULL;
  protected $entity_keys = array();

  // If the entity contains 'created' and 'changed' dates fields.
  protected $date_col = NULL;
  // If the entity contains title field.
  protected $title_col = NULL;
  protected $title_col_name = NULL; // 'title' or 'label' or 'whatyouwant'
  // If the entity contains 'uid' user field.
  protected $user_col = NULL;
  // If the entity contains 'language' language field.
  protected $language_col = NULL;
  // If the entity can be published/unpublished.
  protected $status_col = NULL;
  // If the comment are open or not.
  protected $comment_col = NULL;
  // If the entity can be promoted to front page.
  protected $promote_col = NULL;
  // If the entity can be set as sticky.
  protected $sticky_col = NULL;
  // colomns for bundles.
  protected $name_col = NULL;
  protected $weight_col = NULL;
  protected $description_col = NULL;
  protected $help_col = NULL;
  protected $title_label_col = NULL;
  protected $data_col = NULL;

  // If the entity is fieldable.
  protected $fieldable = NULL;
  // If the entity is exportable.
  protected $exportable = NULL; //TRUE;

  // Correspond to entity_info['bundle of'].
  protected $bundle_of = NULL;
//  Correspond to the bundle entity type.
  /*
   * String ou FAlSE, the bundle entity type. Optional, if another entity is set
   * with 'bundle of' parameter to this entity, the bundle datas will be
   * automatically set.
   */
  protected $has_bundle = NULL;
  // Bundle wildcard.
  protected $menu_wildcard = NULL;

  // the primary key.
  protected $primary_key = NULL;
  protected $tablename = NULL;
  protected $tablename_revision = NULL;
  protected $entityclassname = NULL;
  protected $controller_class = NULL;
  protected $ui_controller_class = NULL;
  protected $entity_translation_class = NULL;
  protected $entity_access_callback = NULL;
  protected $entity_label_callback = NULL;
  /*
   * Class for building entity_info(). Optional, it will be set from the object.
   */
  protected $info_class = NULL;
  protected $install_class = NULL;
  protected $source_class = NULL;
  /*
   * Add key-value to $info['customizable entities']
   */
  protected $customizable_entities_entry = array();

  // Permissions
  protected $create = NULL;
  protected $update = NULL;
  protected $view = NULL;
  protected $delete = NULL;

  // add view modes.
  protected $view_modes = array();

  // wrappers
  protected $views_controller_class = NULL; //'EntityDefaultViewsController';
  protected $rules_controller_class = NULL; //'EntityDefaultRulesController';
  protected $metadata_controller_class = NULL; //'CustomizableEntitiesMetadataController'; //'EntityDefaultMetadataController';
  // Bellow : need some improvement, with this function, the label inline/above
  // on manage display fields does not appear.
  protected $extra_fields_controller_class = NULL; //'CustomizableEntitiesExtraFieldsController';

  // In order to avoid conflict with core or contrib modules, reject thoses entitynames.
  protected $prohibed_entitynames = array(
    // drupal core
    'comment',
    'file',
    'node',
    'taxonomy_term',
    'taxonomy_vocabulary',
    'user',
    // drupal contrib
    'commerce_customer_profile',
    'commerce_line_item',
    'commerce_order',
    'commerce_order_type',
    'commerce_payment_transaction',
    'commerce_product',
    'rules_config',
    'relation',
    'og_membership',
    'og_membership_type',
    'registration',
    'registration_state',
    'registration_type',
    'message',
    'message_type',
    'message_type_category',
    'Workflow',
    'WorkflowConfigTransition',
    'WorkflowScheduledTransition',
    'WorkflowState',
    'WorkflowTransition',
    'better_watchdog_ui_watchdog',
  );

  /*
   * Boolean, shortcut for automatically set $this datas corresponding for same
   * properties as a node.
   */
  protected $same_as_node = NULL;
  /*
   * Boolean, shortcut for automatically set $this datas corresponding for same
   * properties as a node.
   */
  protected $same_as_content_type = NULL;


  /**
   * Set the new object.
   */
  public function __construct($mymodule = NULL, $source = NULL, $parent_class = NULL, $source_class = NULL) {

    if (!empty($source)) {
      // rebuild $this parameters from $source
      $this->setup($source);
    }

    if ($this->entityname === NULL || preg_match('#[^a-z0-9_]+#', $this->entityname)) {
      throw new Exception('A valid entity name is needed in order to create an instance of CustomizableEntitiesEntityInfo class.');
    }
    if (in_array($this->entityname, $this->prohibed_entitynames)) {
      throw new Exception('The entity name ' . $this->entityname . ' is not authorized in order to create an instance of CustomizableEntitiesEntityInfo class.');
    }
    if (empty($this->label)) {
      $this->label = $this->entityname;
    }
    if (empty($this->plural_label)) {
      $this->plural_label = $this->label;
    }
    // Provide default values :
    if (!empty($mymodule)) {
      $this->module = $mymodule;
    }
    elseif ($this->module === NULL) {
      $this->module = 'customizable_entities';
    }

    if ($this->source_class === NULL && !empty($source_class) && class_exists($source_class)) {
      $this->source_class = $source_class;
    }

    if ($this->same_as_node) {
      if ($this->primary_key === NULL) {
        $this->primary_key = 'nid';
      }
      if ($this->revision === NULL) {
        $this->revision = TRUE;
      }
      if ($this->fieldable === NULL) {
        $this->fieldable = TRUE;
      }
      if ($this->title_col === NULL) {
        $this->title_col = TRUE;
      }
      if ($this->title_col_name === NULL) {
        $this->title_col_name = 'title';
      }
      if ($this->date_col === NULL) {
        $this->date_col = TRUE;
      }
      if ($this->user_col === NULL) {
        $this->user_col = TRUE;
      }
      if ($this->status_col === NULL) {
        $this->status_col = TRUE;
      }
      if ($this->promote_col === NULL) {
        $this->promote_col = TRUE;
      }
      if ($this->sticky_col === NULL) {
        $this->sticky_col = TRUE;
      }
      if ($this->comment_col === NULL) {
        $this->comment_col = TRUE;
      }
    }

    if ($this->same_as_content_type) {
      if ($this->revision === NULL) {
        $this->revision = FALSE;
      }
      if ($this->fieldable === NULL) {
        $this->fieldable = FALSE;
      }
      if ($this->title_col === NULL) {
        $this->title_col = TRUE;
      }
      if ($this->title_col_name === NULL) {
        $this->title_col_name = 'label';
      }
      if ($this->name_col === NULL) {
        $this->name_col = TRUE;
      }
      if ($this->help_col === NULL) {
        $this->help_col = TRUE;
      }
      if ($this->description_col === NULL) {
        $this->description_col = TRUE;
      }
      if ($this->title_label_col === NULL) {
        $this->title_label_col = TRUE;
      }
      if ($this->view === NULL) {
        $this->view = FALSE;
        $this->update = TRUE;
      }
    }

    if (empty($this->primary_key)) {
      $this->primary_key = 'id';
    }
    if ($this->entityclassname === NULL) {
      $this->entityclassname = 'CustomizableEntitiesEntity';
    }
    if ($this->controller_class === NULL) {
      $this->controller_class = 'CustomizableEntitiesEntityController';
    }
    if ($this->ui_controller_class === NULL) {
      $this->ui_controller_class = 'CustomizableEntitiesEntityUIController';
    }
    if (empty($this->info_class)) {
      if (!empty($parent_class) && class_exists($parent_class)) {
        $this->info_class = $parent_class;
      }
      else {
        $get_class = get_class($this);
        $this->info_class = $get_class ? $get_class : 'CustomizableEntitiesEntityInfo';
      }
    }
    if (empty($this->install_class)) {
      $this->install_class = 'CustomizableEntitiesInstall';
    }
    if ($this->entity_access_callback === NULL) {
      $this->entity_access_callback = 'customizable_entities_entity_entity_access';
    }
    if ($this->views_controller_class === NULL) {
      $this->views_controller_class = 'EntityDefaultViewsController';
    }
    if ($this->rules_controller_class === NULL) {
      $this->rules_controller_class = 'EntityDefaultRulesController';
    }
    if ($this->metadata_controller_class === NULL) {
      $this->metadata_controller_class = 'CustomizableEntitiesMetadataController';
    }
    if ($this->extra_fields_controller_class === NULL) {
      $this->extra_fields_controller_class = 'CustomizableEntitiesExtraFieldsController';
    }
    if ($this->entity_translation_class === NULL) {
      $this->entity_translation_class = 'EntityTranslationDefaultHandler';
    }

    if (!empty($this->exportable)) {
      $this->entity_keys['status'] = 'export_status';
      $this->entity_keys['module'] = 'export_module';
    }

    if (empty($this->menu_wildcard)) {
      $this->menu_wildcard = ($this->bundle_of) ? '%customizable_entities_bundle_object' : '%customizable_entities_entity_object';
    }

    $entityname_url = str_replace('_', '-', $this->entityname);
    $empty_crud_path = FALSE;
    if (empty($this->crud_path)) {
      $empty_crud_path = TRUE;
      $this->crud_path = $entityname_url;
    }
    if (empty($this->root_admin_ui)) {
      $this->root_admin_ui = "admin/structure/customizable-entities/{$entityname_url}";
    }

    // Automatic config for bundles and associated entity.
    if ($this->bundle_of) {
      // This is the bundle entity.
      $this->name_col = TRUE;
      $this->title_col = TRUE;
      if (empty($this->title_col_name)) {
        $this->title_col_name = "label";
      }
      if ($this->description_col === NULL) {
        $this->description_col = TRUE;
      }
      if ($this->help_col === NULL) {
        $this->help_col = TRUE;
      }
      // Content type has no view page, only edit page.
      //$this->perm_view = FALSE;
      // For Bundle : crud into admin page.
      // Set myentity/add, myentity/edit, etc... instead of
      // admin/structure/myentity/add, admin/structure/myentity/edit, etc...
      if ($empty_crud_path) {
        $this->crud_path = $this->root_admin_ui . '/manage';
      }
      if (empty($this->overview_path)) {
        $this->overview_path = $this->root_admin_ui;
      }
    }

    if (!empty($this->title_col) && empty($this->title_col_name)) {
      $this->title_col_name = "title";
    }

    if (empty($this->overview_path)) {
      $this->overview_path = 'admin/content/' . $this->crud_path;
    }

    if ($this->create === NULL) {
      $this->create = TRUE;
    }
    if ($this->update === NULL) {
      $this->update = TRUE;
    }
    if ($this->view === NULL) {
      $this->view = TRUE;
    }
    if ($this->delete === NULL) {
      $this->delete = TRUE;
    }

    if (empty($this->view_modes)) {
      $this->view_modes = array(
        'full' => array(
          'label' => t('Full'),
          //'custom settings' => FALSE,
        ),
        'teaser' => array(
          'label' => t('Teaser'),
          //'custom settings' => TRUE,
        ),
      );
    }

    // Not implemented yet.
    if (!empty($this->comment_col)) {
      $this->comment_col = FALSE;
    }

    if (empty($this->tablename)) {
      $this->tablename = $this->entityname;
    }
    if (empty($this->tablename_revision)) {
      $this->tablename_revision = $this->entityname . '_revision';
    }

    // Handle possible conflict with existing tables.
    // It happens if a user give (as an example) 'system' as entity_name.
    // The entity table will be system_ce1 and system_ce1_revisions.
    $i = 0;
    $static_name = $this->tablename;
    do {
      $new_tablename = empty($i) ? $static_name : $static_name . '_ce' . $i;
      $not_entity_table = FALSE;
      if ($table_exists = db_table_exists($new_tablename)) {
        // The entity table already exists. It could be the entity table OR an
        // existing tables with name conflict.
        $stored_fields = customizable_entities_retreive_cols_from_db($new_tablename);
        if (empty($stored_fields)) {
          // Fails to retreive the fields. It may not be a mysql installation.
          // If this is a drupal table, we should get it :
          if ($schema = drupal_get_schema($new_tablename)) {
            $stored_fields = $schema['fields'];
          }
        }
        // Export_status ans export_module are quite specific of the
        // customizable_entities module. if present Assume that the table is
        // the entity table
        $not_entity_table = (!isset($stored_fields[$this->primary_key]) || !isset($stored_fields['export_status']) || !isset($stored_fields['export_status']));
      }
      $conflict_table = $table_exists && $not_entity_table;
      $i++;
    } while ($conflict_table);
    $this->tablename = $new_tablename;
    $this->tablename_revision = $this->tablename . '_revision';
  }

  protected function setup($source) {
    // Do something : modify $this datas depending on $source object, as an example,
    // set $this->tablename dynamically.
    // If we want to skip a source, set $this->entityname to ''.
  }

  /**
   * Inform the system about one or more entity types
   *
   * Building entity_info corresponding to settings above.
   *
   * @see hook_entity_info()
   */
  public function entity_info() {

    if (empty($this->entityname)) {
      return array();
    }
    $return = array();

    // Create the entity name :
    $path_admin = "{$this->root_admin_ui}";

    $entity_info = array(
      'label' => $this->label,
      'plural label' => $this->plural_label,
      'entity class' => $this->entityclassname,
      'module' => $this->module,
      'uri callback' => 'entity_class_uri',
      'translation' => array(
        'locale' => TRUE,
      ),
      'base table' => $this->tablename,
    );
    if (class_exists($this->controller_class)) {
      $entity_info['controller class'] = $this->controller_class;
    }

    if ($this->revision) {
      $entity_info['revision table'] = $this->tablename_revision;
    }

    // Build the 'entity keys' according to the setting of each entity.
    $entity_keys = $this->entity_keys;
    if (!empty($this->primary_key)) {
      $entity_keys['id'] = $this->primary_key;
    }
    if ($this->revision) {
      $entity_keys['revision'] = 'vid';
    }
    if ($this->language_col) {
      $entity_keys['language'] = 'language';
    }
    if ($this->name_col) {
      $entity_keys['name'] = 'name';
    }
    if ($this->has_bundle) {
      $entity_keys['bundle'] = 'type';
    }
    if ($this->title_col) {
      $entity_keys['label'] = $this->title_col_name;
    }
    $entity_info['entity keys'] = $entity_keys;

    // If fieldable.
    $entity_info['fieldable'] = $this->fieldable;
    $entity_info['exportable'] = $this->exportable;
    if (!empty($this->entity_access_callback) && function_exists($this->entity_access_callback)) {
      $entity_info['access callback'] = $this->entity_access_callback;
    }
    if ($this->bundle_of) {
      $entity_info['bundle of'] = $this->bundle_of;
    }

    // Add a custom entries : 'root crud path' for the root path for editing, viewing, etc...
    $crud = array('create', 'view', 'update', 'delete');
    $crud_infos = array();
    foreach ($crud as $perm) {
      if (!empty($this->$perm)) {
        $crud_infos[$perm] = TRUE;
      }
    }

    $entity_info['customizable entities'] = array(
      'info class' => $this->info_class,
      'install class' => $this->install_class,
      'crud path' => $this->crud_path,
      'crud' => $crud_infos,
      'overview path' => $this->overview_path,
    ) + $this->customizable_entities_entry;

    if (!empty($this->source_class)) {
      $entity_info['customizable entities']['source class'] = $this->source_class;
    }
    // Set the declared fields in order to inform the install class to create thoses fields.

    $cols = array(
      'primary_key' => array($this->primary_key),
      'date_col' => array('created', 'changed'),
      'title_col' => array($this->title_col_name),
      'user_col' => array('uid'),
      'language_col' => array('language'),
      'status_col' => array('status'),
      'comment_col' => array('comment'),
      'promote_col' => array('promote'),
      'sticky_col' => array('sticky'),
      'weight_col' => array('weight'),
      'description_col' => array('description'),
      'help_col' => array('help'),
      'title_label_col' => array('title_label'),
      'data_col' => array('data'),
    );
    $sql_cols = array();
    foreach ($cols as $key => $schema_names) {
      if ($this->$key) {
        foreach ($schema_names as $schema_name) {
          $sql_cols[] = $schema_name;
        }
      }
    }
    // Below : this will be overwritten by drupal_schema_fields_sql) in
    // entity_get_info(). Only used for the install class to add appropriate columns.
    $entity_info['schema_fields_sql']['base table'] = $sql_cols;
    // Enable the entity API's admin UI.
    $entity_info['admin ui']['path'] = $path_admin;
    $entity_info['admin ui']['controller class'] = $this->ui_controller_class;
    if (!empty($this->menu_wildcard)) {
      $entity_info['admin ui']['menu wildcard'] = $this->menu_wildcard;
    }

    if (!empty($this->entity_label_callback)) {
      // Set dynamic titles.
      $entity_info['label callback'] = $this->entity_label_callback;
    }

    // Manage bundles.
    $entity_info['bundles'] = array();
    // Ih has_bundle is not specified but this entity is register as "bundle of"
    // another entity, thoses parameters will be set into the
    // customizable_entities_build_entity_infos() function.
    if ($this->has_bundle) {
      $entity_info['bundle keys'] = array('bundle' => 'name');
      $entity_info['bundles'] = array();
      // We need another entity info to buils this : the bundles key will be
      // populate into the customizable_entities_entity_info_alter() function.
    }
    elseif ($this->fieldable) {
      // Create 1 bundle : to allow to add fields to if fiedable is TRUE.
      $entity_info['bundles'][$this->entityname] = array(
        'label' => 'Manage ' . $this->plural_label,
        'admin' => array(
          'path' => $path_admin,
          'access arguments' => array(customizable_entities_permission_item('administer', $this->entityname)),
        ),
      );
    }

    $entity_info['view modes'] = $this->view_modes;

    // classes.
    if (!empty($this->views_controller_class) && class_exists($this->views_controller_class)) {
      $entity_info['views controller class'] = $this->views_controller_class;
    }
    if (!empty($this->rules_controller_class) && class_exists($this->rules_controller_class)) {
      $entity_info['rules controller class'] = $this->rules_controller_class;
    }
    if (!empty($this->metadata_controller_class) && class_exists($this->metadata_controller_class)) {
      $entity_info['metadata controller class'] = $this->metadata_controller_class;
    }
    if (!empty($this->extra_fields_controller_class) && class_exists($this->extra_fields_controller_class)) {
      $entity_info['extra fields controller class'] = $this->extra_fields_controller_class;
    }

    // Handle the entity_translation module.
    if (module_exists('entity_translation')) {
      $entity_info['translation']['entity_translation']['class'] = $this->entity_translation_class;
      // Provide correct path
      $delta = 'default';
      $et = array();
      // 'base path', 'view path', 'edit path', 'translate path', 'path wildcard', 'admin theme'
      $wildcard = (!empty($this->menu_wildcard)) ? $this->menu_wildcard : '%customizable_entities_entity_object';
      $et['path schemes'][$delta]['base path'] = "{$entity_info['customizable entities']['crud path']}/{$wildcard}";
      $et['path schemes'][$delta]['path wildcard'] = $wildcard;
      $entity_info['translation']['entity_translation'] += $et;
    }

    // Set the array.
    $return[$this->entityname] = $entity_info;
    return $return;
  }
}

/**
 * Create multiples entities, one entity per provided source.
 *
 * Classes that extends this class must provide in the __construct() method
 * $this->sources parameter. This parameter can be generated by another entity,
 * just set $this->source_class.
 * The info_class must extend the CustomizableEntitiesEntityInfo class or its childs and this
 * class must provide a setup($source) method that will alter $this datas from
 * $source ($source is the value of the array $this->sources). This destination
 * class must provide a __construct() method to. See CustomizableEntitiesEntityInfoCustomizableEntitiesUiBuilded
 * (destination class) for an example.
 */
class CustomizableEntitiesEntityInfoSubentities {
  protected $info_class = '';
  // Set new variables for retrieving parameters
  protected $source_class = NULL;
  protected $source_entity_info = array();
  // An array of the source class objects keyed by their id.
  protected $sources = array();
  protected $module = '';

  public function __construct($mymodule = NULL) {
    if (empty($this->info_class) || !class_exists($this->info_class)) {
      throw new Exception('Cannot create an instance of CustomizableEntitiesEntityInfoSubentities without a valid info class.');
    }
    if (!empty($mymodule)) {
      $this->module = $mymodule;
    }

    if (!empty($this->source_class)) {
      if (empty($this->source_entity_info) && class_exists($this->source_class)) {
        $class = $this->source_class;
        $source_entity_info = customizable_entities_call_method($class, 'entity_info');
        if (!empty($source_entity_info) && count($source_entity_info) == 1) {
          reset($source_entity_info);
          $this->source_entity_info = current($source_entity_info);
        }
      }
      if (!empty($this->source_entity_info)) {
        if (!empty($this->source_entity_info['base table']) && db_table_exists($this->source_entity_info['base table'])) {
          // retreive the builder
          $query = db_query('SELECT * FROM {' . $this->source_entity_info['base table'] . '} ORDER by ' . $this->source_entity_info['entity keys']['id']);
          while ($row = $query->fetchObject()) {
            // Unserialize datas if needed.
            if (isset($row->data) && !empty($row->data)) {
              $row->data = unserialize($row->data);
            }
            $this->sources[$row->{$this->source_entity_info['entity keys']['id']}] = $row;
          }
        }
      }

      if (empty($this->module) && !empty($this->source_entity_info['module'])) {
        $this->module = $this->source_entity_info['module'];
      }
    }
    else {
      // $this->source_class is empty.
      if (empty($this->sources)) {
        throw new Exception('Cannot create an instance of ' . $this->info_class . ' without a specified source.');
      }
    }
  }

  /**
   * Build entity infos.
   *
   *   @see hook_entity_info()
   */
  public function entity_info($provided_source = NULL) {
    $entities_infos = array();
    // Create one entity per source entity.
    $info_class = $this->info_class;
    if (class_exists($info_class)) {
      $parent_class = get_class($this);
      $iterate = array();
      if (!empty($provided_source)) {
        $iterate[] = $provided_source;
      }
      else {
        $iterate = $this->sources;
      }
      foreach ($iterate as $id => $source) {
        $method = 'entity_info';
        $args = array($this->module, $source, $parent_class, $this->source_class);
        $entity_entity_info = customizable_entities_call_method($info_class, $method, $args);
        if (!empty($entity_entity_info) && is_array($entity_entity_info)) {
          // Add a reference to the source entity, not only the class.
          reset($entity_entity_info);
          $entity_type = key($entity_entity_info);
          $entity_entity_info[$entity_type]['customizable entities']['source entity id'] = $id;
          $entities_infos = array_merge($entities_infos, $entity_entity_info);
        }
      }
    }
    return $entities_infos;
  }
}

/**
 * Handle database changes and permissions according to a defined entity_info.
 */
class CustomizableEntitiesInstall {

  protected $entity_info = array();
  protected $entity_infos = array();
  protected $entity_type = '';
  protected $ce_fields = array();

  /**
   * Construct datas for this install class.
   *
   * The main trick of this class is to be able to build schema before the core
   * entity_infos have been build. this is relaisezd by provideding the
   * entity_info on creating this class, throught the construct method.
   */
  public function __construct($entity_type, $entity_infos) {
    if (empty($entity_type) || !is_string($entity_type)) {
      throw new Exception('Cannot create an instance of CustomizableEntitiesInstall without a specified entity type.');
    }
    $this->entity_type = $entity_type;
    if (!empty($entity_infos) && !is_array($entity_infos)) {
      throw new Exception('Cannot create an instance of CustomizableEntitiesInstall without specified entity infos.');
    }
    if (isset($entity_infos[$entity_type])) {
      $this->entity_infos = $entity_infos;
      $this->entity_info = $entity_infos[$entity_type];
      $this->ce_fields = cutomizable_entities_defined_properties($this->entity_info);
    }

  }
  /**
   * Define the current version of the database schema (hook_schema()).
   *
   * Building schema corresponding to settings above.
   *
   * @return
   *   An array, the tables schema keyed by the table names.
   *
   *   @see hook_schema()
   */
  public function schema() {

    if (empty($this->entity_info)) {
      return array();
    }
    $schema = array();
    $entity_table = $this->entity_info['base table'];
    $schema[$entity_table] = $this->schema_dp_table(FALSE);

    $schema[$entity_table]['fields'] = array_merge($this->schema_dp_fields(FALSE), $this->schema_custom_fields(FALSE));

    // Create the revision table.
    if (!empty($this->entity_info['entity keys']['revision'])) {
      $revision_table = $this->entity_info['revision table'];
      $schema[$revision_table] = $this->schema_dp_table(TRUE);
      $schema[$revision_table]['fields'] = array_merge($this->schema_dp_fields(TRUE), $this->schema_custom_fields(TRUE));
    }

    return $schema;
  }

  /**
   * Provide schema table informations.
   *
   * Especially indexes, table description. Fields are defined outside from here.
   *
   * @return
   *   An array, the table infos keyed by properties ('description', etc...).
   *
   * @see CustomizableEntitiesEntityInfo::schema()
   */
  protected function schema_dp_table($revision = FALSE) {

    $label = $this->entity_info['label'];
    $schema = array(
      'description' => (!$revision) ? "Stores {$label} datas" : "Stores {$label} revisions datas",
    );

    if (!isset($schema['primary key'])) {
      $schema['primary key'] = array((!$revision) ? $this->entity_info['entity keys']['id'] : 'vid');
    }

    if ($revision) {
      $schema['unique keys']['vid'] = array('vid');
    }

    // Foreigner keys.
    if (in_array('uid', $this->entity_info['schema_fields_sql']['base table'])) {
      $user_fk = (!$revision) ? 'entity_author' : 'version_author';
      $field_fk = (!$revision) ? 'uid' : 'revision_uid';
      $schema['foreign keys'][$user_fk] = array(
        'table' => 'users',
        'columns' => array($field_fk => 'uid'),
      );
    }
    // The foreigner key for the bundle will be added into the customizable_entities_schema_alter() function.

    if (!empty($this->entity_info['entity keys']['revision'])) {
      $rev_fk = (!$revision) ? 'entity_revision' : 'versioned_entity';
      $col_fk = (!$revision) ? 'vid' : $this->entity_info['entity keys']['id'];
      $schema['foreign keys'][$rev_fk] = array(
        'table' => (!$revision) ? $this->entity_info['revision table'] : $this->entity_info['base table'],
        'columns' => array($col_fk => $col_fk),
      );
    }

    // indexes.
    if (!empty($this->entity_info['entity keys']['bundle']) && !$revision) {
      $schema['indexes']['type'] = array('type');
    }
    if (in_array('uid', $this->entity_info['schema_fields_sql']['base table'])) {
      $schema['indexes']['uid'] = (!$revision) ? array('uid') : array('revision_uid');
    }
    if ($revision) {
      $schema['indexes'][$this->entity_info['entity keys']['id']] = array($this->entity_info['entity keys']['id']);
    }
    else {
      $date_col = (in_array('date', $this->entity_info['schema_fields_sql']['base table']) || in_array('created', $this->entity_info['schema_fields_sql']['base table']));
      if ($date_col) {
        $schema['indexes']['entity_changed'] = array('changed');
        $schema['indexes']['entity_created'] = array('created');
      }
      if (in_array('language', $this->entity_info['schema_fields_sql']['base table'])) {
        $schema['indexes']['language'] = array('language');
      }
      if (in_array('promote', $this->entity_info['schema_fields_sql']['base table']) && in_array('status', $this->entity_info['schema_fields_sql']['base table']) && in_array('sticky', $this->entity_info['schema_fields_sql']['base table']) && $date_col) {
        $schema['indexes']['entity_frontpage'] = array('promote', 'status', 'sticky', 'created');
      }
      // TODO : add :
      // 'node_status_type'    => array('status', 'type', 'nid'),
      //'node_title_type'     => array('title', array('type', 4)),
      //'node_type'           => array(array('type', 4)),
      //'tnid'
    }

    $schema['module'] = $this->entity_info['module'];

    return $schema;
  }

  /**
   * Define schema for DP fields.
   *
   * DP fields are 'uid', 'created', 'changed', 'vid', etc...
   *
   * @param $revision
   *   A Boolean, TRUE for revision table fields. Default to FALSE.
   *
   * @return
   *   An array, the DP schema fields keyed by there name.
   *
   * @see CustomizableEntitiesEntityInfo::schema()
   */
  protected function schema_dp_fields($revision = FALSE) {
    $label = $this->entity_info['label'];
    // Fix primary key.
    $colomns = array();
    if (!$revision) {
      $colomns[$this->entity_info['entity keys']['id']] = array(
        'description' => "Primary Key: Unique {$label} identifier.",
        'type' => 'serial',
        'unsigned' => TRUE,
        'not null' => TRUE,
      );
    }
    else {
      $colomns[$this->entity_info['entity keys']['id']] = array(
        'description' => 'The {' . $this->entity_info['base table'] . '} this version belongs to.',
        'type' => 'int',
        'unsigned' => TRUE,
        'not null' => TRUE,
        'default' => 0,
      );
    }

    if (!empty($this->entity_info['entity keys']['revision'])) {
      if (!$revision) {
        $colomns['vid'] = array(
          'description' => 'The current {' . $this->entity_info['revision table'] . '}.vid version identifier.',
          'type' => 'int',
          'unsigned' => TRUE,
          'not null' => FALSE,
          'default' => NULL,
        );
      }
      else {
        $colomns['vid'] = array(
          'type' => 'serial',
          'unsigned' => TRUE,
          'not null' => TRUE,
          'description' => 'The primary identifier for this revision.',
        );
      }
    }

    if (!empty($this->entity_info['entity keys']['name'])) {
      $colomns[$this->entity_info['entity keys']['name']] = array(
        'description' => "Unique identifier, the machine-readable name of this {$label}.",
        'type' => 'varchar',
        'length' => 255,
        'not null' => TRUE,
      );
    }
    if (!empty($this->entity_info['entity keys']['bundle'])) {
      // For node table, type is not stored into the revision table.
      if (!$revision) {
        $colomns[$this->entity_info['entity keys']['bundle']] = array(
          'description' => "The machine-readable name of this {$label}.",
          'type' => 'varchar',
          'length' => 255,
          'not null' => TRUE,
        );
      }
    }

    if (in_array('description', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['description'] = array(
        'description' => "A brief description of this {$label}.",
        'type' => 'text',
        'not null' => TRUE,
        'size' => 'medium',
        'translatable' => TRUE,
      );
    }

    if (in_array('help', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['help'] = array(
        'description' => "Help information shown to the user when creating a content of this {$label}.",
        'type' => 'text',
        'not null' => TRUE,
        'size' => 'medium',
        'translatable' => TRUE,
      );
    }
    // Set user field
    if (in_array('uid', $this->entity_info['schema_fields_sql']['base table'])) {
      if (!$revision) {
        $colomns['uid'] = array(
          'description' => "The {users}.uid that owns this {$label}; initially, this is the user that created it.",
          'type' => 'int',
          'not null' => TRUE,
          'default' => 0,
        );
      }
      else {
        $colomns['revision_uid'] = array(
          'description' => 'The {users}.uid that created this version.',
          'type' => 'int',
          'not null' => TRUE,
          'default' => 0,
        );
      }
    }

    $date_col = (in_array('date', $this->entity_info['schema_fields_sql']['base table']) || in_array('created', $this->entity_info['schema_fields_sql']['base table']));
    if ($date_col) {
      if (!$revision) {
        $colomns['created'] = array(
          'description' => "The Unix timestamp when the {$label} was created.",
          'type' => 'int',
          'not null' => TRUE,
          'default' => 0,
        );
        $colomns['changed'] = array(
          'description' => "The Unix timestamp when the {$label} was most recently saved.",
          'type' => 'int',
          'not null' => TRUE,
          'default' => 0,
        );
      }
      else {
        $colomns['timestamp'] = array(
          'description' => "The Unix timestamp when this version was created.",
          'type' => 'int',
          'not null' => TRUE,
          'default' => 0,
        );
      }
    }

    if (in_array('language', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['language'] = array(
        'description' => "The Drupal {languages}.language of this {$label}.",
        'type' => 'varchar',
        'length' => 12,
        'not null' => TRUE,
        'default' => '',
      );
    }

    if (!empty($this->entity_info['entity keys']['label'])) {
      $colomns[$this->entity_info['entity keys']['label']] = array(
        'description' => (!$revision) ? "The title of this {$label}, always treated as non-markup plain text." : 'The title of this version.',
        'type' => 'varchar',
        'length' => 255,
        'not null' => TRUE,
        'default' => '',
      );
    }

    $rev_data = ($revision) ? ' (at the time of this revision)' : '';

    if (in_array('weight', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['weight'] = array(
        'type' => 'int',
        'not null' => TRUE,
        'default' => 0,
        'size' => 'tiny',
        'description' => "The weight of this {$label} in relation to others.",
      );
    }

    if ($revision) {
      $colomns['log'] = array(
        'description' => 'The log entry explaining the changes in this version.',
        'type' => 'text',
        'not null' => TRUE,
        'size' => 'big',
      );
    }

    if (in_array('comment', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['comment'] = array(
        'description' => "Whether comments are allowed on this {$label}" . $rev_data . ': 0 = no, 1 = closed (read only), 2 = open (read/write).',
        'type' => 'int',
        'not null' => TRUE,
        'default' => 0,
      );
    }

    if (in_array('status', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['status'] = array(
        'description' => "Boolean indicating whether the {$label}" . $rev_data . ' is published (visible to non-administrators).',
        'type' => 'int',
        'not null' => TRUE,
        'default' => 1,
      );
    }

    if (in_array('promote', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['promote'] = array(
        'description' => "'Boolean indicating whether the {$label}" . $rev_data . ' should be displayed on the front page.',
        'type' => 'int',
        'not null' => TRUE,
        'default' => 0,
      );
    }

    if (in_array('sticky', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['sticky'] = array(
        'description' => "Boolean indicating whether the {$label}" . $rev_data . ' should be displayed at the top of lists in which it appears.',
        'type' => 'int',
        'not null' => TRUE,
        'default' => 0,
      );
    }

    if (in_array('title_label', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['title_label'] = array(
        'description' => 'The label displayed for the title field on the edit form.',
        'type' => 'varchar',
        'length' => 255,
        'not null' => TRUE,
        'default' => '',
        'translatable' => TRUE,
      );
    }

    if (in_array('data', $this->entity_info['schema_fields_sql']['base table'])) {
      $colomns['data'] = array(
        'description' => "The parameters of this {$label}.",
        'type' => 'blob',
        'not null' => FALSE,
        'size' => 'big',
        'serialize' => TRUE,
      );
    }

    // If I want to make survey entity exportable I need to add thoses fields.
    // Edit the entity keys for changing the default export status name, to
    // avoid conflict with the status colomn.
    $colomns = $colomns + entity_exportable_schema_fields('export_module', 'export_status');

    // @TODO : add tnid and translate ?

    return $colomns;
  }

  /**
   * Define schema for custom fields.
   *
   * each class should implement it.
   *
   * @return
   *   An array, the custom schema fields or an empty array if there is
   *   no fields.
   *
   * @see CustomizableEntitiesEntityInfo::schema()
   */
  protected function schema_custom_fields($revision = FALSE) {
    return array();
  }

  /**
   * Update schema fields regarding changes on the LS database.
   *
   * This method is called on saving a survey entity, compare the current schema
   * stored into a variable and add colomns if needed. Set a watchdog entry if
   * changes are made.
   *
   * @return
   *   A boolean, TRUE if the schema has been updated (so the cached schema need
   *   to be rebuild).
   *
   * @see CustomizableEntitiesEntityInfo::schema()
   */
  public function update_table($reset_cache = TRUE) {

    if (empty($this->entity_info)) {
      return NULL;
    }
    // Test the base table first, then the revision table.
    $base_table = $this->entity_info['base table'];
    $revision_table = (!empty($this->entity_info['revision table'])) ? $this->entity_info['revision table'] : "{$base_table}_revision";
    $iterate = array($base_table);
    $iterate[] = $revision_table;

    $schema = $this->schema();
    $changes = array();
    $table_revision_created = FALSE;
    foreach ($iterate as $tablename) {
      $update_fields = FALSE;
      $revision_loop = ($tablename != $base_table);
      $table_should_exists = (!$revision_loop || (!empty($this->entity_info['entity keys']['revision'])));

      $info_schema = (!empty($schema[$tablename])) ? $schema[$tablename] : array();
      if (!$table_should_exists && db_table_exists($tablename)) {
        // The revision table but it should not. we delete it
        db_drop_table($tablename);
        $changes['table']['deleted'][] = $tablename;
      }
      elseif ($table_should_exists && !db_table_exists($tablename)) {
        // Table does not exists : save it.
        db_create_table($tablename, $info_schema);
        // save it into a variable.
        $changes['table']['created'][] = $tablename;
        if ($revision_loop) {
          $table_revision_created = TRUE;
        }
      }
      elseif ($revision_loop && empty($this->entity_info['entity keys']['revision'])) {
        continue;
      }
      else {
        // The table exists : look for changes on fields.
        $variable_name = "customizable_entities_{$tablename}_fields";
        $stored_fields = customizable_entities_retreive_cols_from_db($tablename);
        if (!$stored_fields) {
          $stored_fields_name = variable_get($variable_name, array());
        }
        else {
          $stored_fields_name = array_keys($stored_fields);
        }
        $schema_fields = (!empty($schema[$tablename]['fields'])) ? $schema[$tablename]['fields'] : array();
        $schema_fields_name = array_keys($schema_fields);
        $diff = array_diff($schema_fields_name, $stored_fields_name);
        if (!empty($diff)) {
          // Somes colomns have been added from the entityInfo class.
          foreach ($diff as $field) {
            // for revision, the bundle col type exists
            if (!db_field_exists($tablename, $field) && !empty($info_schema['fields'][$field])) {
              db_add_field($tablename, $field, $info_schema['fields'][$field]);
              $changes['field'][$tablename]['created'][] = $field;
              $update_fields = TRUE;
            }
          }
        }
        // Check for deleted fields on from the entity info class :
        $diff_del = array_diff($stored_fields_name, $schema_fields_name);
        if (!empty($diff_del)) {
          // Somes colomns have been deleted from entity info class.
          foreach ($diff_del as $field) {
            if (db_field_exists($tablename, $field)) {
              // Delete this colomn only if there all values are all NULL into DP.
              $query = db_select($tablename, 't');
              $query->addField('t', $field);
              $result = $query->execute();
              $field_values = $result->fetchCol();
              $values = array_values(array_unique($field_values));
              if (count($values) < 1 || (count($values) == 1 && (empty($values[0]) || (isset($stored_fields[$field]['default']) && $values[0] == $stored_fields[$field]['default'])))) {
                db_drop_field($tablename, $field);
                $changes['field'][$tablename]['deleted'][] = $field;
                $update_fields = TRUE;
              }
            }
          }
        }

      }

      if ($update_fields && !empty($variable_name) && !empty($schema_fields_name)) {
        // Save the new schema into a variable. on vairiable per table (so 2
        // variables if revision enabled).
        variable_set($variable_name, $schema_fields_name);
      }
    }

    // Tables have been created or updated.
    if (!empty($changes)) {
      // watchdog the changes
      $actions = array('created', 'deleted');
      foreach ($actions as $action) {
        if (!empty($changes['table'][$action])) {
          foreach ($changes['table'][$action] as $tablename) {
            watchdog('customizable_entities', 'the table %table has been @action.', array(
              '%table' => $tablename,
              '@action' => $action,
            ), WATCHDOG_NOTICE);
          }
        }
        if (!empty($changes['field'])) {
          foreach ($changes['field'] as $tablename => $datas) {
            if (!empty($changes['field'][$tablename][$action])) {
              watchdog('customizable_entities', 'the table %table has been updated, colomns @action : %fields.', array(
                '%table' => $tablename,
                '%fields' => implode(', ', $changes['field'][$tablename][$action]),
                '@action' => $action,
              ), WATCHDOG_NOTICE);
            }
          }
        }
      }

      // If revsion table has been created, synchronize datas between base table
      // and revision table
      if ($table_revision_created) {
        // Find field of revision to populate from base
        $base_fields = array_keys($schema[$base_table]['fields']);
        $rev_fields = array_keys($schema[$revision_table]['fields']);
        $source2rev_fields = array_intersect($base_fields, $rev_fields);
        // Retreive base table ID name
        $base_field_name = FALSE;
        foreach ($schema[$base_table]['fields'] as $field => $schema_field) {
          if ($schema_field['type'] == 'serial') {
            $base_field_name = $field;
            break;
          }
        }
        if ($base_field_name && !empty($source2rev_fields)) {
          $sources = db_select($base_table, 'b')
            ->fields('b', $source2rev_fields)
            ->condition('vid', NULL)
            //->condition('b.uid', 0, '!=')
            ->execute()
            ->fetchAll();
          foreach ($sources as $source) {
            $update = array();
            foreach ($source2rev_fields as $field) {
              $update[$field] = $source->$field;
            }
            // Log doesn't have a default value :
            $update['log'] = '';
            db_insert($revision_table)
              ->fields($update)
              ->execute();
          }

          $revisions = db_select($revision_table, 'r')
            ->fields('r', array($base_field_name, 'vid'))
            ->execute()
            ->fetchAll();
          // This foreach updates each user only once as for each user there
          // can be only 1 record in the {users} table with vid = 0.
          foreach ($revisions as $revision) {
            db_update($base_table)
              ->condition($base_field_name, $revision->$base_field_name)
              ->condition('vid', NULL)
              ->fields(array('vid' => $revision->vid))
              ->execute();
          }
        }
      }

      // Reset cache.
      if ($reset_cache) {
        drupal_get_schema(NULL, TRUE);
      }
    }
    return $changes;
  }

  /**
   * Provide hook_permission for the entity.
   *
   * @return
   *   An array, the entity permissions.
   *
   * @see hook_permission()
   */
  public function permission($datas = array()) {

    if (empty($this->entity_info)) {
      return array();
    }
    // We set up permisssions to manage entity types, manage all entities and the
    // permissions for each individual entity
    // @TODO : need to translate this throw an handler.
    $arg = array(
      '@type' => $this->entity_info['label'],
      '@types' => $this->entity_info['plural label'],
    );

    $has_author = in_array('uid', $this->entity_info['schema_fields_sql']['base table']);

    $admin_perm = customizable_entities_permission_item('administer', $this->entity_type);
    $permissions = array(
      $admin_perm => array(
        'title' => t('Administer @types', $arg),
        'description' => t('Create and delete @types, and set their permissions.', $arg),
        'restrict access' => TRUE,
      ),
    );

    if (!empty($this->entity_info['customizable entities']['overview path'])) {
      $perm_overview = customizable_entities_permission_item('overview', $this->entity_type);
      $permissions += array(
        $perm_overview => array(
          'title' => t('Access the @types overview page', $arg),
          'description' => user_access($perm_overview)
            ? t('Get an overview of <a href="@url">all @types</a>.', array_merge($arg, array('@url' => url($this->entity_info['customizable entities']['overview path']))))
            : t('Get an overview of all @types.', $arg),
        ),
      );
    }

    // Manage bundles.
    $has_bundle = ($this->entity_info['entity keys']['bundle']);

    // Handle bundles for permissions.
    foreach ($this->entity_info['bundles'] as $name => $bundle) {

      $perm_human_prefix =  ($has_bundle) ? t('%type_name:', array('%type_name' => $bundle['label'])) . ' ' :  '';

      if (!empty($this->entity_info['customizable entities']['crud']['create'])) {
        $permissions += array(
          customizable_entities_permission_item('add', $this->entity_type, $name) => array(
            'title' => $perm_human_prefix . t('Create new @type', $arg),
          ),
        );
      }

      // Update permission.
      if (!empty($this->entity_info['customizable entities']['crud']['update'])) {
        if ($has_author) {
          $permissions += array(
            customizable_entities_permission_item('update own', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('Edit own @type', $arg),
            ),
            customizable_entities_permission_item('update any', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('Edit any @type', $arg),
            ),
          );
        }
        else {
          $permissions += array(
            customizable_entities_permission_item('update', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('Edit @type', $arg),
            ),
          );
        }
      }

      // View permission.
      if (!empty($this->entity_info['customizable entities']['crud']['view'])) {
        if ($has_author) {
          $permissions += array(
            customizable_entities_permission_item('view own', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('View own @type', $arg),
            ),
            customizable_entities_permission_item('view any', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('View any @type', $arg),
            ),
          );
        }
        else {
          $permissions += array(
            customizable_entities_permission_item('view', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('View @type', $arg),
            ),
          );
        }
      }

      // Delete permission.
      if (!empty($this->entity_info['customizable entities']['crud']['delete'])) {
        if ($has_author) {
          $permissions += array(
            customizable_entities_permission_item('delete own', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('Delete own @type', $arg),
            ),
            customizable_entities_permission_item('delete any', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('Delete any @type', $arg),
            ),
          );
        }
        else {
          $permissions += array(
            customizable_entities_permission_item('delete', $this->entity_type, $name) => array(
              'title' => $perm_human_prefix . t('Delete @type', $arg),
            ),
          );
        }
      }

    }

    if (!empty($this->entity_info['customizable entities']['crud']['view']) && in_array('status', $this->entity_info['schema_fields_sql']['base table'])) {
      if ($has_author) {
        $permissions += array(
          customizable_entities_permission_item('view own unpublished', $this->entity_type) => array(
            'title' => t('view own unpublished @type', $arg),
          ),
        );
      }
      else {
        $permissions += array(
          customizable_entities_permission_item('view unpublished', $this->entity_type) => array(
            'title' => t('view unpublished @type', $arg),
          ),
        );
      }
    }

    if (!empty($this->entity_info['entity keys']['revision'])) {
      // Handle revision permissions. No bundle consideration.
      $permissions += array(
        customizable_entities_permission_item('view revisions', $this->entity_type) => array(
          'title' => t('View any @type revisions', $arg),
        ),
        customizable_entities_permission_item('revert revisions', $this->entity_type) => array(
          'title' => t('Revert any @type revisions', $arg),
        ),
        customizable_entities_permission_item('delete revisions', $this->entity_type) => array(
          'title' => t('Delete any @type revisions', $arg),
        ),
      );
    }

    // Custom permissions
    $custom_permissions = $this->permission_custom($datas);
    $permissions = array_merge($permissions, $custom_permissions);

    return $permissions;
  }

  /**
   * Define custom permissions.
   *
   * each class should implement it.
   *
   * @return
   *   An array, the entity permissions.
   *
   * @see CustomizableEntitiesEntityInfo::permission()
   */
  protected function permission_custom() {
    return array();
  }

  /**
   * Define default bundle settings.
   *
   * Define default settings of entity properties : order on manage fields and
   * manage diplay pages and if the fields should be hidden or visibile by
   * default. Use extra fields defined into the
   * CustomizableEntitiesExtraFieldsController class. Weight of each property
   * can be set into the bundle_settings_default_order() method.
   *
   * @return
   *   nothing.
   *
   * @see CustomizableEntitiesExtraFieldsController::fieldExtraFields()
   */
  public function bundle_settings_default($extra) {
    if (empty($extra) || !is_array($extra)) {
      return;
    }

    if (!empty($this->entity_info['bundles'])) {
      foreach ($this->entity_info['bundles'] as $bundle => $datas) {
        $variable_name = 'field_bundle_settings_' . $this->entity_type . '__' . $bundle;
        $field_bundle_settings = variable_get($variable_name, array());
        if (empty($field_bundle_settings)) {
          // The variable has never been set. Set default values.
          $settings = array(
            'view_modes' => array(),
            'extra_fields' => array(
              'form' => array(),
              'display' => array(),
            ),
          );
          $order = $this->bundle_settings_default_order($extra);

          if (!empty($order)) {
            $extra_fields_form = (!empty($extra[$this->entity_type][$bundle]['form'])) ? $extra[$this->entity_type][$bundle]['form'] : array();
            $extra_fields_display = (!empty($extra[$this->entity_type][$bundle]['display'])) ? $extra[$this->entity_type][$bundle]['display'] : array();
            $weight = (count($order) + 5) * -1;
            $i = 0;
            foreach ($order as $field_name => $display) {
              if (isset($extra_fields_form[$field_name])) {
                // This extra field exists on 'manage fields' page.
                $settings['extra_fields']['form'][$field_name]['weight'] = $weight;
                unset($extra_fields_form[$field_name]);
              }
              if (isset($extra_fields_display[$field_name])) {
                // This extra field exists on 'manage display' page.
                foreach ($display as $view_mode => $visibility) {
                  $settings['extra_fields']['display'][$field_name][$view_mode]['weight'] = $weight;
                  $settings['extra_fields']['display'][$field_name][$view_mode]['visible'] = $visibility;
                  unset($extra_fields_display[$field_name]);
                  // Saves custom view_modes
                  if ($view_mode != 'default') {
                    $i++;
                    $custom_view_modes[$view_mode] = $i;
                  }
                }
              }
              $weight++;
            }
            if (!empty($custom_view_modes)) {
              $settings['view_modes'] = array_flip($custom_view_modes);
            }
            $weight_form = $weight_display = $weight;
            // Add extra fields that have not been declared into order
            foreach ($extra_fields_form as $field_name => $data) {
              $settings['extra_fields']['form'][$field_name]['weight'] = $weight_form;
              $weight_form++;
            }
            foreach ($extra_fields_display as $field_name => $data) {
              $settings['extra_fields']['display'][$field_name][$view_mode]['weight'] = $weight_display;
              $settings['extra_fields']['display'][$field_name][$view_mode]['visible'] = TRUE;
              $weight_display++;
            }
            //$settings['view_modes']['full']['custom_settings'] = FALSE;
            //dpm($variable_name);
            //dpm($settings);
            // save it
            variable_set($variable_name, $settings);
          }
        }
      }
    }
  }

  /**
   * Define order on form and display and visibility.
   *
   * each class should implement it.
   *
   * @return
   *   An array, keyed by the entity colomn (ex : 'uid', 'created', etc...) and
   *   value is an array of a boolean keyed by the view mode : TRUE if visible
   *   or FALSE if hidden for the keyed view mode. The default view mode is
   *   'default'. The order of the returned array determines the weight of each
   *   property.
   *
   * @see CustomizableEntitiesInstall::bundle_settings_default()
   */
  protected function bundle_settings_default_order($extra, $view_mode = 'default') {
    if (empty($this->entity_info)) {
      return array();
    }
    $order = array();

    if (!empty($this->entity_info['entity keys']['id'])) {
      // ID will be not displayed because this extrafields is disabled by default.
      // see : CustomizableEntitiesExtraFieldsController::__construct().
      $order[$this->entity_info['entity keys']['id']] = array($view_mode => TRUE);
    }
    if (!empty($this->entity_info['entity keys']['label'])) {
      $order[$this->entity_info['entity keys']['label']] = array($view_mode => TRUE);
    }
    // Label just before name.
    if (!empty($this->entity_info['entity keys']['name'])) {
      $order[$this->entity_info['entity keys']['name']] = array($view_mode => FALSE);
    }
    if (!empty($this->entity_info['entity keys']['bundle'])) {
      $order[$this->entity_info['entity keys']['bundle']] = array($view_mode => FALSE);
    }
    if (in_array('description', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['description'] = array($view_mode => TRUE);
    }
    if (in_array('help', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['help'] = array($view_mode => FALSE);
    }
    if (in_array('title_label', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['title_label'] = array($view_mode => FALSE);
    }
    if (!empty($this->entity_info['entity keys']['revision'])) {
      $order['log'] = array($view_mode => FALSE);
    }
    if (in_array('uid', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['uid'] = array($view_mode => FALSE);
    }
    if (in_array('created', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['created'] = array($view_mode => FALSE);
    }
    if (in_array('changed', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['changed'] = array($view_mode => FALSE);
    }
    if (in_array('language', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['language'] = array($view_mode => FALSE);
    }
    if (in_array('status', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['status'] = array($view_mode => FALSE);
    }
    if (in_array('comment', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['comment'] = array($view_mode => FALSE);
    }
    if (in_array('promote', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['promote'] = array($view_mode => FALSE);
    }
    if (in_array('sticky', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['sticky'] = array($view_mode => FALSE);
    }
    if (in_array('weight', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['weight'] = array($view_mode => FALSE);
    }
    if (in_array('data', $this->entity_info['schema_fields_sql']['base table'])) {
      $order['data'] = array($view_mode => FALSE);
    }

    return $order;
  }
}
